import os
from unittest import TestCase

import numpy as np
import numpy.testing as test
from .read import _prep

from PyFVCOM.coordinate import utm_from_lonlat
from PyFVCOM.current import *
from PyFVCOM.read import FileReader
from PyFVCOM.utilities.grid import StubFile


class Residuals_test(TestCase):

    def setUp(self):
        self.starttime, self.endtime, self.interval, self.lon, self.lat, self.triangles = _prep()
        self.model = StubFile(self.starttime,
                              self.endtime,
                              self.interval,
                              lon=self.lon,
                              lat=self.lat,
                              triangles=self.triangles,
                              zone='30N')
        self.fvcom = FileReader(self.model.ncfile.name, variables=['u', 'v', 'ua', 'va'])
        self.fvcom.grid.x, self.fvcom.grid.y, _ = utm_from_lonlat(self.fvcom.grid.lon, self.fvcom.grid.lat, zone='30N')
        self.fvcom.grid.xc, self.fvcom.grid.yc, _ = utm_from_lonlat(self.fvcom.grid.lonc, self.fvcom.grid.latc, zone='30N')

    def tearDown(self):
        self.model.ncfile.close()
        os.remove(self.model.ncfile.name)
        del(self.model)

    def test_scalar2vector(self):
        actual_u, actual_v = 1, 1
        test_u, test_v = scalar2vector(45, np.sqrt(2))
        test.assert_almost_equal(test_u, actual_u)
        test.assert_almost_equal(test_v, actual_v)

    def test_vector2scalar(self):
        actual_direction, actual_magnitude = 45, np.sqrt(2)
        test_direction, test_magnitude = vector2scalar(1, 1)
        test.assert_equal(actual_direction, test_direction)
        test.assert_equal(actual_magnitude, test_magnitude)

    def test_depth_resolved_vorticity(self):
        expected_vorticity = np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan,
                                       -1.69406589e-21, np.nan, np.nan, -1.64608942e-22, 6.08804931e-22, np.nan, np.nan,
                                       -6.61744490e-23, -3.97046694e-23, np.nan, np.nan, np.nan, 7.01449159e-22,
                                       1.32348898e-22, np.nan, np.nan, np.nan, -1.19114008e-22, np.nan, np.nan,
                                       4.23516474e-22, 1.32348898e-22, -1.32348898e-23, np.nan, 1.05879118e-22,
                                       4.76456033e-22, np.nan, -2.11758237e-22, np.nan, np.nan, np.nan, np.nan,
                                       2.51462906e-22, np.nan, np.nan, -1.16467030e-21, np.nan, np.nan, np.nan, np.nan,
                                       np.nan, np.nan, -4.76456033e-22, np.nan, np.nan, np.nan, -6.61744490e-24,
                                       2.64697796e-22, np.nan, np.nan, 2.11758237e-22, np.nan, -2.64697796e-23, np.nan,
                                       0.00000000e+00, 5.55865372e-22, -1.58818678e-22, np.nan, np.nan, np.nan, np.nan,
                                       3.17637355e-22, np.nan, np.nan, 4.23516474e-22, np.nan, np.nan, np.nan, np.nan])
        vort = vorticity(self.fvcom)
        # Only check the first time and vertical layer to make my life easier.
        test.assert_almost_equal(expected_vorticity, vort[0, 0, :])

    def test_depth_averaged_vorticity(self):
        expected_vorticity = np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan,
                                       -1.21760986e-21, np.nan, np.nan, -1.05879118e-22, 3.27563523e-22, np.nan,
                                       np.nan, -1.32348898e-23, -4.63221143e-23, np.nan, np.nan, np.nan, 5.95570041e-22,
                                       3.97046694e-22, np.nan, np.nan, np.nan, 0.00000000e+00, np.nan, np.nan,
                                       2.38228016e-22, 2.64697796e-23, -2.97785021e-23, np.nan, 1.05879118e-22,
                                       2.11758237e-22, np.nan, 5.29395592e-23, np.nan, np.nan, np.nan, np.nan,
                                       2.64697796e-23, np.nan, np.nan, -5.82335151e-22, np.nan, np.nan, np.nan, np.nan,
                                       np.nan, np.nan, -1.85288457e-22, np.nan, np.nan, np.nan, -8.27180613e-24,
                                       2.64697796e-23, np.nan, np.nan, 0.00000000e+00, np.nan, -6.61744490e-23, np.nan,
                                       2.64697796e-23, 6.61744490e-23, -1.05879118e-22, np.nan, np.nan, np.nan, np.nan,
                                       2.64697796e-23, np.nan, np.nan, 1.58818678e-22, np.nan, np.nan, np.nan, np.nan])
        vort = vorticity(self.fvcom, depth_averaged=True)
        test.assert_almost_equal(expected_vorticity, vort[0, :])
